! ----------------------------------------------------------------------
! ----------------------------------------------------------------------

! module load intel
! module unload mpich
! module li
! ifort -coarray=shared -coarray-num-images=4 -o Jacobi_CAF Jacobi_CAF.F
! ./Jacobi_CAF

! ----------------------------------------------------------------------

! Write a Jacobi iteration code for the Laplace equation using CAF.

      PROGRAM JACOBI_CAF

      IMPLICIT NONE

      INTEGER NPROC, NR, NC, NCL, MXITER,
     1        me, me_p1, me_m1, nimages,
     1        i, j, jshift, iter

      INTEGER niter[*]
      REAL dt[*]

      PARAMETER ( NPROC = 4, NR = 1000, NC = 1000, NCL = NC/NPROC,
     1            MXITER = 1000)

      REAL T ( 0: NR + 1, 0: NCL + 1), Told ( 0: NR + 1, 0: NCL + 1),
     1     T1d_top ( 0: NR + 1), T1d_botm ( 0: NR + 1), dt_global

      REAL top_send_window ( 0: NR + 1)[*]
      REAL botm_send_window ( 0: NR + 1)[*]

! ----------------------------------------------------------------------
! Figure out which processor "me" is.

! Each image finds out who it is in an ordinal number space of images.
      me = THIS_IMAGE()
      me_m1 = me - 1
      me_p1 = me + 1

! Maximum number of images.
      nimages = NUM_IMAGES()

      IF ( NPROC /= nimages) THEN
        IF ( me == 1) 
     1     WRITE ( 6, *)"Stopping in JACOBI_CAF: NPROC /= nimages."
        ERROR STOP
      END IF


      IF ( me == 1) THEN

        print *, 'Please provide number of iterations to broadcast'
        read *, niter
        niter = MIN0 ( niter, MXITER)
        print *, 'niter = ', niter

! Broadcast "niter" to other images from image # 1.
        DO j = 2, nimages
        niter[j] = niter
        END DO

      END IF

! Ensure that each image got "niter".
      SYNC ALL

! ----------------------------------------------------------------------
! Initialize solution and boundary data.

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        T ( i, j) = 0.0  ! Initialize interior and bdy to 0 to begin with.
        END DO
      END DO

! --------------------

! Now set up right and lower boundaries. Left and upper boundaries are ok as is.

      IF ( me == nimages) THEN
! Right boundary.

      DO i = 0, NR + 1
        T ( i, NCL + 1) = (100.0 / NR) * i
      END DO

      END IF

! ----------

! Lower boundary.

      jshift = NCL * ( me - 1)

      DO j = 0, NCL + 1
        T ( NR + 1, j) = (100.0 / NC) * ( j + jshift)
      END DO

! --------------------

! Copy back to "Told"

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        Told ( i, j) = T ( i, j)
        END DO
      END DO

! ----------------------------------------------------------------------
! Start of iteration loop.

      DO iter = 1, niter

! Ensure that each iteration is synchronized across processors.
      SYNC ALL

! ----------------------------------------------------------------------
! Do a Jacobi iteration here. At the end of this unit, "T" has improved values.

      DO j = 1, NCL
        DO i = 1, NR
          T ( i, j) = 0.25 * ( Told ( i+1, j) + Told ( i-1, j)
     1                       + Told ( i, j+1) + Told ( i, j-1) )
        END DO
      END DO

! ----------------------------------------------------------------------
! Messaging block. Eventually, copy back to "Told" for next Jacobi iteration.

! Put message values in respective p1 and m1 message windows on CAF
! images
      DO i = 0, NR + 1
         botm_send_window ( i) =  T ( i, NCL)
         top_send_window ( i) =  T ( i, 1)
      END DO

! Make sure that "botm_send_window, top_send_window" are loaded on all
! images.
      SYNC ALL

! ------------------------------

      IF ( me /= 1) THEN

! Get message values from right-neighbour window and put in local storage T

        T1d_top ( :) = botm_send_window ( :) [me_m1]

        DO i = 0, NR + 1
          T ( i, 0) = T1d_top ( i)
        END DO

      END IF
      
! Get message values from left-neighbour window and put in local storage T

      IF ( me /= nimages) THEN

        T1d_botm ( :) = top_send_window ( :) [me_p1]

        DO i = 0, NR + 1
          T ( i, NCL + 1) = T1d_botm ( i)
        END DO

      END IF

! Make sure that all data is in the big "T" array.
      SYNC ALL

! ----------------------------------------------------------------------
! Reduction and writing of data done here.

      dt = 0.0

      DO j = 1, NCL
        DO i = 1, NR
          dt = AMAX1 ( ABS ( T ( i, j) - Told ( i, j) ), dt)
        END DO
      END DO

! Make sure that "dt" is built on all images.
      SYNC ALL

      IF ( me == 1) THEN

      dt_global = dt

        DO i = 2, nimages
          dt_global = AMAX1 ( dt_global, dt[i])
        END DO

        WRITE ( 6, *)'iteration number and parallel temp change = ',
     1               iter,' ', dt_global

      END IF

      SYNC ALL

      IF ( ( me == nimages) .AND.
     1     ( MOD ( iter, 10) == 0) ) THEN

      DO i = 9, 0, -1
        print *, T ( NR - i, NCL - i)
      END DO

      END IF

! --------------------
! Now copy back to "Told".

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        Told ( i, j) = T ( i, j)
        END DO
      END DO

! ----------------------------------------------------------------------

      END DO

! End of iteration loop.
! ----------------------------------------------------------------------

      END PROGRAM JACOBI_CAF

! ----------------------------------------------------------------------
! ----------------------------------------------------------------------


