! ----------------------------------------------------------------------
! ----------------------------------------------------------------------

! module load intel
! module load mpich
! module li
! mpif90 -o Jacobi_onesided Jacobi_onesided_Get.F
! mpiexec -n 4 ./Jacobi_onesided

! ----------------------------------------------------------------------

! Write a Jacobi iteration code for the Laplace equation using MPI.

      PROGRAM JACOBI_MPI

      IMPLICIT NONE

      include 'mpif.h'

      INTEGER NPROC, NR, NC, NCL, MXITER,
     1        my_rank, ierr, my_rank_m1, my_rank_p1, numprocs,
     1        tag, i, j, jshift, iter, niter, nr_p2

      PARAMETER ( NPROC = 4, NR = 1000, NC = 1000, NCL = NC/NPROC,
     1            MXITER = 1000)

      INTEGER status(MPI_STATUS_SIZE)

      REAL T ( 0: NR + 1, 0: NCL + 1), Told ( 0: NR + 1, 0: NCL + 1),
     1     T1d_top ( 0: NR + 1), T1d_botm ( 0: NR + 1), dt, dt_global

      INTEGER(KIND=MPI_ADDRESS_KIND) :: window_size, window_offset

      INTEGER :: type_mpi_real, top_window_handle, botm_window_handle

      REAL top_send_window(0: NR +1), botm_send_window( 0: NR+1)

! ----------------------------------------------------------------------
! Set-up MPI. The MPI_Bcast must be seen by all processors.

      CALL MPI_INIT ( ierr)

      CALL MPI_COMM_RANK ( MPI_COMM_WORLD, my_rank, ierr)

      CALL MPI_COMM_SIZE ( MPI_COMM_WORLD, numprocs, ierr)

      tag = 0

      my_rank_m1 = my_rank - 1

      my_rank_p1 = my_rank + 1

      IF ( my_rank .EQ. 0) THEN

        IF ( numprocs .NE. NPROC) print *,'numprocs .NE. NPROC ' ,
     1       numprocs, ' ', NPROC, ' Think of aborting'

        print *, 'Please provide number of iterations to broadcast'
        read *, niter
        niter = MIN0 ( niter, MXITER)
        print *, 'niter = ', niter

      END IF

      CALL MPI_Bcast ( niter, 1, MPI_INTEGER, 0, MPI_COMM_WORLD, ierr)

! ----------------------------------------------------------------------
! Initialize solution and boundary data.

      nr_p2 = nr + 2 ! Useful for declaring message length in MPI_Send,MPI_Recv

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        T ( i, j) = 0.0  ! Initialize interior and bdy to 0 to begin with.
        END DO
      END DO

! --------------------

! Now set up right and lower boundaries. Left and upper boundaries are ok as is.

      IF ( my_rank .EQ. numprocs - 1) THEN
! Right boundary.

      DO i = 0, NR + 1
        T ( i, NCL + 1) = (100.0 / NR) * i
      END DO

      END IF

! ----------

! Lower boundary.

      jshift = NCL * my_rank

      DO j = 0, NCL + 1
        T ( NR + 1, j) = (100.0 / NC) * ( j + jshift)
      END DO

! --------------------

! Copy back to "Told"

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        Told ( i, j) = T ( i, j)
        END DO
      END DO

! ----------------------------------------------------------------------
! Start of iteration loop.

! Type of variable being called in bytes is stored in "type_mpi_real".
      call MPI_Type_size(MPI_REAL,type_mpi_real,ierr) 

! Calculate window size in bytes.
      window_size = nr_p2 * type_mpi_real

! Create windows for "top_send_window" and "botm_send_window" to hold
! values. "MPI_INFO_NULL" is something that you just do in MPI.
! "top_window_handle" allows us to get a handle on the window. It is
! crucial to release it with "MPI_WIN_FREE". Not freeing it will cause memory
! leaks, even if one is exiting a subroutine.

      call MPI_WIN_CREATE ( top_send_window, window_size, type_mpi_real,
     1     MPI_INFO_NULL, MPI_COMM_WORLD, top_window_handle, ierr)

      call MPI_WIN_CREATE( botm_send_window, window_size, type_mpi_real,
     1     MPI_INFO_NULL, MPI_COMM_WORLD, botm_window_handle, ierr)
      
      top_send_window = 0
      botm_send_window = 0

! ----------------------------------------------------------------------

      DO iter = 1, niter

! Ensure that each iteration is synchronized across processors.
      CALL MPI_Barrier ( MPI_COMM_WORLD, ierr)

! ----------------------------------------------------------------------
! Do a Jacobi iteration here. At the end of this unit, "T" has improved values.

      DO j = 1, NCL
        DO i = 1, NR
          T ( i, j) = 0.25 * ( Told ( i+1, j) + Told ( i-1, j)
     1                       + Told ( i, j+1) + Told ( i, j-1) )
        END DO
      END DO

! ----------------------------------------------------------------------
! Messaging block. Eventually, copy back to "Told" for next Jacobi iteration.

! Put message values in respective p1 and m1 message windows on MPI process
      DO i = 0, NR + 1
         botm_send_window ( i) =  T ( i, NCL)
         top_send_window ( i) =  T ( i, 1)
      END DO

      CALL MPI_Barrier ( MPI_COMM_WORLD, ierr)

! ------------------------------

! Set window offset to 0
      window_offset = 0

      IF ( my_rank .NE. 0) THEN

! With the following call, RMA access epoch starts. Only the window at 
! the processor with rank "my_rank_m1" can be accessed by RMA operations 
! on window "botm_window_handle" after this call. 

! "MPI_LOCK_SHARED" ensures that other processes can also access the
! same "botm_send_window/botm_window_handle" data from the
! processor with "my_rank_m1"

      call MPI_WIN_LOCK ( MPI_LOCK_SHARED, my_rank_m1, 0,
     1                  botm_window_handle, ierr)

! Get message values from right-neighbour window and put in local storage T

         call MPI_GET ( T1d_top, nr_p2, MPI_REAL, my_rank_m1, 
     1        window_offset, nr_p2, MPI_REAL, botm_window_handle, ierr)

! This call completes the RMA access epoch started by "MPI_WIN_LOCK". 
! RMA oprations issued during this period will have completed both at 
! the origin and at the target when the call returns. 

      call MPI_WIN_UNLOCK ( my_rank_m1, botm_window_handle, ierr)

      DO i = 0, NR + 1
        T ( i, 0) = T1d_top ( i)
      END DO

      END IF
      
! Get message values from left-neighbour window and put in local storage T

      IF ( my_rank .NE. numprocs - 1) THEN

! With the following call, RMA access epoch starts. Only the window at 
! the processor with rank "my_rank_p1" can be accessed by RMA operations 
! on window "top_window_handle" after this call. 

      call MPI_WIN_LOCK ( MPI_LOCK_SHARED, my_rank_p1, 0,
     1                  top_window_handle, ierr)

         call MPI_GET ( T1d_botm, nr_p2, MPI_REAL, my_rank_p1, 
     1       window_offset, nr_p2, MPI_REAL, top_window_handle, ierr)
         
! This call completes the RMA access epoch started by "MPI_WIN_LOCK". 
! RMA oprations issued during this period will have completed both at 
! the origin and at the target when the call returns. 

      call MPI_WIN_UNLOCK ( my_rank_p1, top_window_handle, ierr)

      DO i = 0, NR + 1
        T ( i, NCL + 1) = T1d_botm ( i)
      END DO

      END IF

! ----------------------------------------------------------------------
! Reduction and writing of data done here.

      dt = 0.0

      DO j = 1, NCL
        DO i = 1, NR
          dt = AMAX1 ( ABS ( T ( i, j) - Told ( i, j) ), dt)
        END DO
      END DO

      dt_global = 0.0

      CALL MPI_Reduce ( dt, dt_global, 1, MPI_REAL, MPI_MAX, 0,
     1                  MPI_COMM_WORLD, ierr)

      IF ( my_rank == 0) THEN
        print *,'iteration number and parallel temp change = ',
     1          iter,' ', dt_global

      END IF

! This barrier needed if we want the serial convergence before the trace.
      CALL MPI_Barrier ( MPI_COMM_WORLD, ierr)

      IF ( ( my_rank == numprocs - 1) .AND.
     1     ( MOD ( iter, 10) == 0) ) THEN

      DO i = 9, 0, -1
        print *, T ( NR - i, NCL - i)
      END DO

      END IF

! --------------------
! Now copy back to "Told".

      DO j = 0, NCL + 1
        DO i = 0, NR + 1
        Told ( i, j) = T ( i, j)
        END DO
      END DO

! ----------------------------------------------------------------------

      END DO

! End of iteration loop.
! ----------------------------------------------------------------------

      call MPI_WIN_FREE ( top_window_handle, ierr)

      call MPI_WIN_FREE ( botm_window_handle, ierr)

      CALL MPI_Finalize (ierr)

! ----------------------------------------------------------------------

      END PROGRAM JACOBI_MPI

! ----------------------------------------------------------------------
! ----------------------------------------------------------------------


